mod checkbox;

use std::env;
use std::sync::Arc;

use lettre::message::{header::ContentType, Message, Mailbox, MultiPart};
use lettre::transport::{smtp::authentication::Mechanism, smtp::authentication::Credentials};
use lettre::transport::smtp::AsyncSmtpTransport;
use lettre::AsyncTransport;
use serde::{Serialize, Deserialize};
use askama::{Template};
use sqlx::{FromRow, query_as, postgres::{PgPool, PgQueryResult, PgPoolOptions},  query, query_scalar};
use axum::{Form, response::IntoResponse, routing::get, Router, body::Body, response::Response};
use axum::extract::State;
use axum_csrf::{CsrfConfig, CsrfLayer, CsrfToken, CsrfError};

#[derive(Serialize, sqlx::Type, Deserialize, Debug, Clone, Copy)]
#[serde(rename_all = "lowercase")]
#[sqlx(type_name = "campus")]
enum Campus {
	#[sqlx(rename = "c")]
	Calgary,
	#[sqlx(rename = "l")]
	Lethbridge,
}

#[derive(Template, Deserialize, Serialize)]
#[template(path = "course_interest.html")]
struct FormPage {
	authentication_key: String,
}

#[derive(Template, Deserialize, Serialize)]
#[template(path = "thank_you.html")]
struct InterestPage {
	compiler_design: i64,
	distributed_systems: i64,
	calgary_cpsc_group: i64,
}

#[derive(Debug, FromRow)]
struct DBFormResponse {
	name: String,
	email: String,
	uleth_id: i32,
	compiler_design: bool,
	distributed_systems: bool,
	calgary_cpsc_group: bool,
	school_campus: Campus,
	comments: String,
}
impl From<HTMLFormResponse> for DBFormResponse {
	fn from(html_form: HTMLFormResponse) -> Self {
		DBFormResponse {
			name: html_form.name,
			email: html_form.email,
			uleth_id: html_form.uleth_id,
			compiler_design: html_form.compiler_design,
			distributed_systems: html_form.distributed_systems,
			calgary_cpsc_group: html_form.calgary_cpsc_group,
			school_campus: html_form.school_campus,
			comments: html_form.comments,
		}
	}
}

#[derive(Debug, Serialize, Deserialize, Clone)]
struct HTMLFormResponse {
	name: String,
	email: String,
	uleth_id: i32,
	#[serde(default = "bool::default")]
	#[serde(deserialize_with = "checkbox::deserialize_checkbox")]
	compiler_design: bool,
	#[serde(default = "bool::default")]
	#[serde(deserialize_with = "checkbox::deserialize_checkbox")]
	distributed_systems: bool,
	#[serde(default = "bool::default")]
	#[serde(deserialize_with = "checkbox::deserialize_checkbox")]
	calgary_cpsc_group: bool,
	comments: String,
	school_campus: Campus,
	authentication_key: String,
}

struct Interest {
	distributed_systems: Option<i64>,
	compiler_design: Option<i64>,
	calgary_cpsc_group: Option<i64>,
}

impl DBFormResponse {
	async fn insert(pool: &PgPool, to_add: &Self) -> Result<PgQueryResult, sqlx::Error> {
		query!(
			r#"
INSERT INTO responses
	(name,email,uleth_id,compiler_design,distributed_systems,calgary_cpsc_group,comments,school_campus)
VALUES
	($1, $2, $3, $4, $5, $6, $7, $8)
			"#,
			to_add.name,
			to_add.email,
			to_add.uleth_id,
			to_add.compiler_design,
			to_add.distributed_systems,
			to_add.calgary_cpsc_group,
			to_add.comments,
			to_add.school_campus as Campus)
			.execute(&*pool)
			.await
	}
	async fn all(pool: &PgPool) -> Result<Vec<DBFormResponse>, sqlx::Error> {
		query_as!(DBFormResponse,
			r#"
SELECT name,email,uleth_id,compiler_design,distributed_systems,calgary_cpsc_group,school_campus AS "school_campus: Campus",comments
FROM responses;
			"#)
			.fetch_all(&*pool)
			.await
	}
	async fn count(pool: &PgPool) -> Result<Option<i64>, sqlx::Error> {
		query_scalar!(
			r#"
SELECT COUNT(id)
FROM responses;
			"#)
			.fetch_one(&*pool)
			.await
	}
	async fn interest_counts(pool: &PgPool) -> Result<Interest, sqlx::Error> {
		query_as!(Interest,
			r#"
SELECT
	SUM(CASE WHEN distributed_systems = true THEN 1 ELSE 0 END) AS distributed_systems,
	SUM(CASE WHEN compiler_design = true THEN 1 ELSE 0 END) AS compiler_design,
	SUM(CASE WHEN calgary_cpsc_group = true THEN 1 ELSE 0 END) AS calgary_cpsc_group
FROM responses;
			"#)
			.fetch_one(&*pool)
			.await
	}
}

async fn index(
	token: CsrfToken,
	State(ServerState { db }): State<ServerState>,
) -> Result<(CsrfToken, FormPage), FormError> {
	let tmpl = FormPage {
		authentication_key: token.authenticity_token().unwrap(),
	};
	Ok((token, tmpl))
}

#[derive(Debug, Serialize, Deserialize)]
pub enum FormError {
	EmailError(String),
	DatabaseError(String),
	FormValidationError(String),
}
impl IntoResponse for FormError {
	fn into_response(self) -> Response<Body> {
		match self {
			Self::EmailError(s) => {
				eprintln!("{}", s);
				"There was an error sending the email to your inbox."
			},
			Self::DatabaseError(s) => {
				eprintln!("{}", s);
				"There was an error saving your response to the database."
			},
			Self::FormValidationError(s) => {
				eprintln!("{}", s);
				"There was an error processing the form; it should be resolved by going back, refreshing the page, then submitting again."
			},
		}.to_string().into_response()
	}
}

impl From<CsrfError> for FormError {
	fn from(e: CsrfError) -> Self {
		Self::FormValidationError(format!("{:?}", e))
	}
}
impl From<lettre::address::AddressError> for FormError {
	fn from(e: lettre::address::AddressError) -> Self {
		Self::EmailError(format!("{:?}", e))
	}
}
impl From<lettre::transport::smtp::Error> for FormError {
	fn from(e: lettre::transport::smtp::Error) -> Self {
		Self::EmailError(format!("{:?}", e))
	}
}
impl From<lettre::error::Error> for FormError {
	fn from(e: lettre::error::Error) -> Self {
		Self::EmailError(format!("{:?}", e))
	}
}
impl From<sqlx::Error> for FormError {
	fn from(e: sqlx::Error) -> Self {
		Self::DatabaseError(format!("{:?}", e))
	}
}

async fn check_key(
	token: CsrfToken,
	State(ServerState { db }): State<ServerState>,
	Form(payload): Form<HTMLFormResponse>,
) -> Result<InterestPage, FormError> {
    // Verfiy the Hash and return the String message.
    token.verify(&payload.authentication_key)?;
		let db_form_resposne = payload.clone().into();
		let _ = DBFormResponse::insert(&db, &db_form_resposne).await?;
		let smtp_host = env::var("SMTP_HOST").unwrap();
		let smtp_user = env::var("SMTP_USER").unwrap();
		let smtp_pass = env::var("SMTP_PASS").unwrap();
		let msg = Message::builder()
			.from("Tait Hoyem <tait@tait.tech>".parse()?)
			.to(Mailbox::new(Some(payload.name.clone()), payload.email.clone().parse()?))
			.bcc("Tait Hoyem <tait.hoyem@uleth.ca>".parse()?)
			.subject("Thank Your For Your Interest")
			.multipart(MultiPart::alternative_plain_html(
				String::from("Thank you for registering you interest in one of: \r\n\r\n- Compiler Design\r\n- Distributed Systems\r\n- Calgary-based computer-science group\r\n\r\nWe will keep you informed of the progress as we receive more signatories. You will only be contacted if there is an update regarding a thing you have registered interest in.\r\n\r\nIf you'd like to opt out, please let me, tait@tait.tech know."),
				String::from("<p>Thank you for registering you interest in one of:</p> <ul><li>Compiler Design</li><li>Distributed Systems</li><li>Calgary-based computer-science group</li></ul><p>We will keep you informed of the progress as we receive more signatories. You will only be contacted if there is an update regarding a thing you have registered interest in.</p><p>If you'd like to opt out, please let me, tait@tait.tech know.</p>"),
			))?;
		let sender = AsyncSmtpTransport::<lettre::Tokio1Executor>::starttls_relay(&smtp_host)?
			.credentials(Credentials::new(
				smtp_user,
				smtp_pass,
			))
			.authentication(vec![Mechanism::Login])
			.build();
		let result = sender.send(msg).await;
		let db_standings = DBFormResponse::interest_counts(&db).await?;
		let standings = InterestPage {
			calgary_cpsc_group: db_standings.calgary_cpsc_group.unwrap_or(0i64),
			compiler_design: db_standings.compiler_design.unwrap_or(0i64),
			distributed_systems: db_standings.distributed_systems.unwrap_or(0i64),
		};
		Ok(standings)
}

#[derive(Clone)]
struct ServerState {
	db: Arc<PgPool>,
}

async fn db_connect() -> PgPool {
	PgPoolOptions::new()
		.max_connections(8)
		.connect(&env::var("DATABASE_URL").unwrap())
		.await
		.unwrap()
}

#[tokio::main]
async fn main() {
	let config = CsrfConfig::default();
	let db = db_connect().await;
	let state = ServerState {
		db: Arc::new(db),
	};
	let app = Router::new()
		.route("/", get(index).post(check_key))
		.layer(CsrfLayer::new(config))
		.with_state(state);
	let listener = tokio::net::TcpListener::bind("0.0.0.0:3000").await.unwrap();
	axum::serve(listener, app).await.unwrap();
}
